# coding=UTF-8
"""
UNIX things
"""
from __future__ import print_function, absolute_import, division

import logging
import os

from .daemon import daemonize
from .pidlock import AcquirePIDLock
from .signals import hang_until_sig

logger = logging.getLogger(__name__)


def is_running_as_root():
    """
    Is this process running as root?

    Checks whether EUID is 0
    :return: bool
    """
    return os.geteuid() == 0



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import logging
import os
import sys

import six

try:
    import pwd
    import grp
except ImportError:
    # Windows?
    class L(object):    # always return UID=GID=0
        def __getattr__(self, item):
            return lambda q: 0
    pwd = grp = L()

from satella.coding import typed, Callable

logger = logging.getLogger(__name__)

DEVNULL = '/dev/null'


@typed(Callable, bool, (None, int), (None, int))
def daemonize(exit_via=sys.exit,
              redirect_std_to_devnull=True,
              uid=None, gid=None):
    """
    Make this process into a daemon.

    This entails:

    - umask 0
    - forks twice
    - be the child of init
    - becomes session leader
    - changes root directory to /
    - closes stdin, stdout, stderr
    - (option) redirects stdin, stdout, stderr to /dev/null

    Refer - "Advanced Programming in the UNIX Environment" 13.3

    :param exit_via: callable used to terminate process
    :param redirect_std_to_devnull: whether to redirect stdin, stdout and
        stderr to /dev/null
    :param uid: User to set (via seteuid). Default - this won't be done. You
        can pass either user name as string or UID.
    :param gid: Same as UID, but for groups. These will be resolved too.
    :raises KeyError: uid/gid was passed as string, but getpwnam() failed
    """

    _double_fork(exit_via=exit_via)
    _close_descriptors()

    if redirect_std_to_devnull:
        _redirect_descriptors_to_null()

    _parse_ug(uid, pwd, 'pw_uid', os.seteuid)
    _parse_ug(gid, grp, 'gr_gid', os.setegid)


def _parse_ug(no, module, fieldname, osfun):
    if no is not None:
        if isinstance(no, six.string_types):
            no = getattr(module.getpwnam(no), fieldname)
        osfun(no)


def _redirect_descriptors_to_null():
    sys.stdin = open(DEVNULL, 'rb')
    sys.stdout = open(DEVNULL, 'wb')
    sys.stderr = open(DEVNULL, 'wb')


def _close_descriptors():
    for d in [sys.stdin, sys.stdout, sys.stderr]:
        d.close()


def _double_fork(exit_via):
    os.umask(0)

    if os.fork() > 0:
        exit_via()  # parent exits

    os.setsid()

    if os.fork() > 0:
        exit_via()  # parent exits
    os.chdir('/')



# coding=UTF-8
"""
Handle signals, like a boss
"""
from __future__ import print_function, absolute_import, division

import logging
import signal
import time

from satella.coding import typed

logger = logging.getLogger(__name__)

end = False


def __sighandler(a, b):
    global end
    end = True


@typed(list)
def hang_until_sig(extra_signals=[]):
    """Will hang until this process receives SIGTERM or SIGINT.
    If you pass extra signal IDs (signal.SIG*) with extra_signals,
    then also on those signals this call will release."""
    global end

    # Ascertain what Python are we working on. 2012 PyPy and earlier
    # may be affected by https://bugs.pypy.org/issue1255

    bugged_pypy = False
    try:
        import platform
    except:
        pass
    else:
        if platform.python_implementation() == 'PyPy':
            try:
                mon, day, year = platform.python_build()[1].split(' ')
                year = int(year)
            except:
                pass
            else:
                bugged_pypy = year <= 2012

    signal.signal(signal.SIGTERM, __sighandler)
    signal.signal(signal.SIGINT, __sighandler)
    for s in extra_signals:
        signal.signal(s, __sighandler)

    while not end:
        try:
            if bugged_pypy:
                time.sleep(1)  # see https://bugs.pypy.org/issue1255
            else:
                signal.pause()
        except:  # pause() is undefined on Windows
            try:  # we will sleep for small periods of time
                time.sleep(0.5)
            except IOError:  # "Interrupted system call"
                pass



# coding=UTF-8
"""
Just useful objects to make your coding nicer every day
"""
from __future__ import print_function, absolute_import, division

from .concurrent import Monitor, RMonitor, CallableGroup
from .algos import merge_dicts
from .recast_exceptions import rethrow_as, silence_excs
from .typecheck import typed, List, Tuple, Dict, NewType, Callable, Sequence, \
    TypeVar, Generic, Mapping, Iterable, Union, Any, Optional, CallSignature, \
    Number, coerce
from .structures import TimeBasedHeap, Heap, typednamedtuple

__all__ = [
    'typednamedtuple',
    'TimeBasedHeap', 'Heap', 'CallableGroup',
    'Monitor', 'RMonitor', 'merge_dicts',
    'typed', 'List', 'Tuple', 'Dict', 'NewType', 'Callable', 'Sequence', 'coerce'
    'TypeVar', 'Generic', 'Mapping', 'Iterable', 'Union', 'Any', 'Optional', 'CallSignature', 'Number',
    'rethrow_as', 'silence_excs'
]




# coding=UTF-8
from __future__ import print_function, absolute_import, division
import six
import logging
import functools

logger = logging.getLogger(__name__)

__all__ = [
    'silence',
    'rethrow_as'
]


def silence_excs(*exc_types):
    """
    Silence given exception types.

    Can be either a decorator or a context manager
    """
    return rethrow_as(*[(t, None) for t in exc_types])


class rethrow_as(object):
    """
    Transform some exceptions into others.

    Either a decorator or a context manager
    """

    def __init__(self, *pairs, **kwargs):
        """
        Pass tuples of (exception to catch - exception to transform to).

        New exception will be created by calling exception to transform to with
        repr of current one.

        You can also provide just two exceptions, eg.

          rethrow_as(NameError, ValueError)

        :param exception_preprocessor: other callable/1 to use instead od repr.
            Should return a text
        """

        # You can also provide just two exceptions
        if len(pairs) == 2 and all(issubclass(p, Exception) for p in pairs):
            a, b = pairs
            pairs = [(a, b)]

        self.to_catch = tuple([p[0] for p in pairs])
        self.pairs = pairs
        self.exception_preprocessor = kwargs.get('exception_preprocessor', repr)

    def __call__(self, fun):
        @functools.wraps(fun)
        def inner(*args, **kwargs):
            with self:
                return fun(*args, **kwargs)
        return inner

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not isinstance(exc_val, self.to_catch):
            return

        for from_, to in self.pairs:
            if isinstance(exc_val, from_):
                if to is None:
                    return True
                else:
                    raise to(self.exception_preprocessor(exc_val))



# coding=UTF-8
"""
Decorator for debug-time typechecking
"""
from __future__ import print_function, absolute_import, division

import inspect
import itertools
import logging

import six
from copy import copy
try:
    import typing
except ImportError:
    from backports import typing
from collections import namedtuple
import functools
import itertools

logger = logging.getLogger(__name__)

List = typing.List
Tuple = typing.Tuple
Dict = typing.Dict
NewType = typing.NewType
Callable = typing.Callable
Sequence = typing.Sequence
Number = six.integer_types + (float,)
TypeVar = typing.TypeVar
Generic = typing.Generic
Mapping = typing.Mapping
Iterable = typing.Iterable
Union = typing.Union
Any = typing.Any
Optional = typing.Optional


# Internal tokens - only instances will be
class _NotGiven(object):
    pass


class _NoDefault(object):
    pass


_CSArgument = namedtuple('_CSArgument', ('name', 'required', 'default_value'))


class CSArgument(_CSArgument):
    def __str__(self):
        p = ['Argument ' + self.name]
        if not self.required:
            p.append('optional with default %s' % (self.default_value,))
        return ' '.join(p)

    def __eq__(self, other):
        """
        Is other argument same as this, in name/default value if present?
        :param other: CSArgument
        :return: bool
        """
        return self.name == other.name and \
               self.required == other.required and \
               self.default_value == other.default_value


class CSVarargsPlaceholder(CSArgument):
    pass


class CSKwargsPlaceholder(CSArgument):
    pass


class TypeErrorReason(object):
    pass


class CSTypeError(TypeError):
    """
    A TypeError exception on steroids
    """

    def __str__(self):
        return 'Problem with argument %s' % (arg.name,)

    def __init__(self, arg):
        """
        :param arg: Argument definition
        :type arg: CSArgument
        """
        super(CSTypeError, self).__init__(str(self))
        self.arg = arg


class CSBadTypeError(CSTypeError):
    def __init__(self, arg, expected, got):
        super(CSBadTypeError, self).__init__(arg)

        self.expected = expected
        self.got = got

    def __str__(self):
        return 'Bad type given for arg %s, expected %s got %s' % (
            self.arg.name, self.expected, self.got)


class CSNotGivenError(CSTypeError):
    def __str__(self):
        return 'Value for argument %s not given' % (self.arg.name,)


class CSMultipleValuesGivenError(CSTypeError):
    def __str__(self):
        return 'Got multiple values for argument' % (self.arg.name,)


class CallSignature(object):
    """
    Call signature of a callable.
    
    Properties:
      - has_varargs (Bool) - if has varargs
      - has_kwargs (Bool) - if has kwargs
      - locals (Dict[str => CSArgument]) - list of locals this function call
        will generate
      - pos_args (List[CSArgument)] - list of positional arguments
      - varargs_name (Union[str, None]) - name of varargs argument, or None if
        not present
      - kwargs_name (Union[str, None]) - name of kwargs argument, or None if
        not present
    """

    def count_required_positionals(self):
        c = 0
        for a in self.pos_args:
            if a.required:
                c += 1
        return c

    def __eq__(self, other):
        """
        Compare if two call signatures are IDENTICAL
        :param other: CallSignature
        :return: bool
        """

        if any(a != b for a, b in zip(self.pos_args, other.pos_args)):
            return False

        return self.has_kwargs == other.has_kwargs and \
               self.has_varargs == other.has_varargs

    def __init__(self, callable):
        args, varargs, kwargs, defaults = inspect.getargspec(callable)

        defaults = defaults or ()
        # pad them
        while len(defaults) < len(args):
            defaults = [_NoDefault] + list(defaults)

        # process positionals
        self.pos_args = []
        self.locals = {}
        for arg, default in zip(args, defaults):
            cs_arg = CSArgument(arg,
                                default is _NoDefault,
                                default)
            self.pos_args.append(cs_arg)
            self.locals[arg] = cs_arg

        self.varargs_name = varargs
        if varargs is not None:
            self.has_varargs = True
            self.locals[self.varargs_name] = CSVarargsPlaceholder(
                self.varargs_name, False, [])
        else:
            self.has_varargs = False

        self.kwargs_name = kwargs
        if kwargs is not None:
            self.has_kwargs = True
            self.locals[self.kwargs_name] = CSKwargsPlaceholder(
                self.kwargs_name, False, {})
        else:
            self.has_kwargs = False

    def to_invocation(self, locals):
        """
        Return an invocation to the function reconstructed from its locals
        :param locals: as returned by .result()
        :return: tuple of (args, kwargs)
        """
        locals = copy(locals)
        args = []

        for arg in self.pos_args:
            if arg.name in locals:
                args.append(locals.pop(arg.name))
            elif not arg.required:
                args.append(arg.default_value)
                del locals[arg.name]

        return args, locals

    def result(self, *args, **kwargs):
        """
        Simulate a function call, see what locals are defined
        
        Return a dict of (local_variable_name => it's value),
        or TypeError

        :param args: function call parameters
        :param kwargs: function call parameters
        :return: dict
        :raise CSTypeError: call would raise a TypeError
        """
        assert len(args) >= self.count_required_positionals()

        locals = {}

        # positional
        for arg, value in itertools.izip_longest(self.pos_args,
                                                 args[:len(self.pos_args)],
                                                 fillvalue=_NotGiven):

            if value is _NotGiven:
                if arg.required:
                    raise CSNotGivenError(arg)
                else:
                    value = arg.default_value

            locals[arg.name] = value

        # varargs
        if self.has_varargs:
            locals[self.varargs_name] = args[len(self.pos_args):]

        # kwargs
        if self.has_kwargs:
            locals[self.kwargs_name] = kwargs

        return locals

    def is_match_amount(self, *args, **kwargs):
        """
        Would a function call with these arguments succeed, based solely on
        number and "keywordnessity" or parameters?
        """
        if len(args) > len(self.pos_args):
            if not self.has_varargs:
                return False  # *args expected

        if len(args) < self.count_required_positionals():
            return False  # Not enough posits

        if len(kwargs) > 0 and not self.has_kwargs:
            return False  # kwargs expected

        return True


def __typeinfo_to_tuple_of_types(typeinfo, operator=type):
    if typeinfo == 'self':
        return None
    elif typeinfo is None:
        return (operator(None),)
    elif typeinfo == int and six.PY2:
        return six.integer_types
    else:
        if isinstance(typeinfo, (tuple, list)):
            new_tup = []
            for elem in typeinfo:
                new_tup.extend(__typeinfo_to_tuple_of_types(elem))
            return tuple(new_tup)
        else:
            return (typeinfo,)


def _do_if_not_type(var, type_, fun='default'):

    if type_ in ((type(None), ), ) and (fun == 'default'):
        return None

    if type_ in (None, (None, ), 'self'):
        return var

    if not isinstance(var, type_):

        if fun == 'default':
            if type_[0] == type(None):
                return None
            else:
                print(var, 'to', type_[0])
                return type_[0](var)

        q = fun()
        if isinstance(q, Exception):
            raise q
        return q
    else:
        return var


def typed(*t_args, **t_kwargs):
    """
    Use like:

        @typed(int, six.text_type)
        def display(times, text):
            ...
            
    You can also check for return type with kw argument of "returns", ie.
    
        @typed(int, int, returns=int)
        def sum(a, b):
            return a+b

    Or
        @typed('self', a, b):
        def method(self, a, b):
        ..

    If you specify extra argument - mandatory=True - type will always be
    checked, regardless if debug mode is enabled

    Same rules apply.

    int will automatically include long for checking (Python 3 compatibility)
    If you want to check for None, type (None, )
    None for an argument means "do no checking", (None, ) means "type must be
    NoneType". You can pass tuples or lists to match for multiple types

    :param t_args:
    :param t_kwargs:
    """

    t_args = [(__typeinfo_to_tuple_of_types(x) if x is not None else None)
              for x in t_args]

    t_retarg = t_kwargs.get('returns', None)
    is_mandatory = t_kwargs.get('mandatory', False)

    if t_retarg is not None:
        t_retarg = __typeinfo_to_tuple_of_types(t_retarg)

    def outer(fun):
        if (not __debug__) and (not is_mandatory):
            return fun

        @functools.wraps(fun)
        def inner(*args, **kwargs):
            # add extra 'None' argument if unbound method
            for argument, typedescr in zip(args, t_args):
                _do_if_not_type(argument, typedescr, lambda: \
                    TypeError('Got %s, expected %s' % (
                            type(argument), typedescr)))

            rt = fun(*args, **kwargs)

            return _do_if_not_type(rt, t_retarg, lambda:
                    TypeError('Returned %s, expected %s' % (
                        type(rt), t_retarg)))

        return inner
    return outer


def coerce(*t_args, **t_kwargs):
    """
    Use like:

        @typed(int, six.text_type)
        def display(times, text):
            ...

    You can also check for return type with kw argument of "returns", ie.

        @typed(int, int, returns=int)
        def sum(a, b):
            return a+b

    Or
        @typed('self', a, b):
        def method(self, a, b):
        ..

    If you specify extra argument - mandatory=True - type will always be
    checked, regardless if debug mode is enabled

    Same rules apply.

    int will automatically include long for checking (Python 3 compatibility)
    If you want to check for None, type (None, )
    None for an argument means "do no checking", (None, ) means "type must be
    NoneType". You can pass tuples or lists to match for multiple types

    :param t_args:
    :param t_kwargs:
    """

    t_args = [(__typeinfo_to_tuple_of_types(x, operator=lambda x: x))
              for x in t_args]

    def argify(args):
        return [_do_if_not_type(argument, typedescr) \
                        for argument, typedescr in six.moves.zip_longest(args, t_args)]

    t_retarg = t_kwargs.get('returns', None)

    t_retarg = __typeinfo_to_tuple_of_types(t_retarg, operator=lambda x: x)

    def outer(fun):

        @functools.wraps(fun)
        def inner(*args, **kwargs):
            # add extra 'None' argument if unbound method

            new_args = argify(args)

            rt = fun(*new_args, **kwargs)
            return _do_if_not_type(rt, t_retarg)

        return inner

    return outer



# coding=UTF-8
from __future__ import print_function, absolute_import, division
import six
import logging
from ..typecheck import typed, Callable

logger = logging.getLogger(__name__)


__all__ = [
    'CallableGroup',
]

class CallableGroup(object):
    """
    This behaves like a function, but allows to add other functions to call
    when invoked, eg.

        c1 = Callable()

        c1.add(foo)
        c1.add(bar)

        c1(2, 3)

    Now both foo and bar will be called with arguments (2, 3). Their exceptions
    will be propagated.

    """

    # todo not threadsafe with oneshots

    def __init__(self, gather=False, swallow_exceptions=False):
        """
        :param gather: if True, results from all callables will be gathered
                       into a list and returned from __call__
        :param swallow_exceptions: if True, exceptions from callables will be
                                   silently ignored. If gather is set,
                                   result will be the exception instance
        """
        self.callables = []  # tuple of (callable, oneshot)
        self.gather = gather
        self.swallow_exceptions = swallow_exceptions

    @typed(None, Callable, bool)
    def add(self, callable, oneshot=False):
        """
        :param oneshot: if True, callable will be unregistered after single call
        """
        self.callables.append((callable, oneshot))

    def __call__(self, *args, **kwargs):
        """
        Run the callable. All registered callables will be called with
        passed arguments, so they should have the same arity.

        If callables raise, it will be passed through.

        :return: list of results if gather was set, else None
        """
        clbl = self.callables  # for moar thread safety
        self.callables = []

        if self.gather:
            results = []

        for callable, oneshot in clbl:
            try:
                q = callable(*args, **kwargs)
            except Exception as e:
                if not self.swallow_exceptions:
                    raise  # re-raise
                q = e

            if self.gather:
                results.append(q)

            if not oneshot:
                self.callables.append((callable, oneshot))

        if self.gather:
            return results


